
-include $(TOP)/common/gmsl

Q=@
SIMUL=1

AVR_MCU=ATMEGA128

ifdef VARIANT
  $(eval $(shell grep '^'$(VARIANT)';' options.txt | cut -d';' -f 2))
  $(eval $(shell grep '^'$(VARIANT)';' options.txt | cut -d';' -f 3))
  $(eval $(shell grep '^'$(VARIANT)';' options.txt | cut -d';' -f 4))
endif

ifeq ($(SIMUL),0)
  override SIMUL=
endif

COMMON_SRC_DIR=$(TOP)/common
COMMON_INC_DIRS=$(TOP) $(TOP)/common $(TOP)/platform

DEF += GAGE_CORE=gage_core_$(CORE_SIZE) CRYPTO_BYTES=32

OBJ_DIR = obj
LST_DIR = lst

PROG      = main.elf

ARCH=AVR

_PREFIX=
_FLAGS=

INTEL_PREFIX=
INTEL_FLAGS=


AVR_PREFIX=avr-
MCU:=$($(ARCH)_MCU)

CORE_OBJ := $(addprefix $(OBJ_DIR)/,$(CORE_OBJ))

AFLAGS =-no-pad-sections $(addprefix -I,$(COMMON_INC_DIRS)) -I. -adhlmns=$(LST_DIR)/$(notdir $(@:.o=.lst)) --defsym MEASURE_ASM=0 \
  $(foreach def,$(ADEF), --defsym $(def))


AVR_FLAGS=$(addprefix -I,$(COMMON_INC_DIRS)) -I. -mmcu=$(call lc,$(MCU)) -D$(call uc,$(MCU)) -Dinline= -DF_CPU=16000000 \
 -ffunction-sections -fdata-sections \
 -Wa,-adhlns=$(LST_DIR)/$(notdir $(@:.o=.lst))
 
AVR_LD_FLAGS= -Wl,-Map=main.map,--cref \
 -Wl,--relax,--gc-sections \
 -Wl,--undefined=_mmcu,--section-start=.mmcu=0x91000 \
 -Wl,--gc-sections -lc 

ARM_PREFIX=arm-none-eabi-
ARM_FLAGS=

M0_PREFIX=$(ARM_PREFIX)
M0_FLAGS=-mtune=cortex-m0

M3_PREFIX=$(ARM_PREFIX)
M3_FLAGS=-mtune=cortex-m3

M4_PREFIX=$(ARM_PREFIX)
M4_FLAGS=-mtune=cortex-m4

MSP430_PREFIX=msp430-
MSP430_FLAGS=

SIMULATOR=simavr-ecl.elf

PLATFORM_OBJ=avr_setup.o circularbytebuffer-asm.o avr_uart_i-asm.o performance_test.o

PROGRAMMER     = avrispmkII
PROG_PORT      = usb

OBJ       = $(OBJ_DIR)/main.o $(OBJ_DIR)/gage_std.o $(OBJ_DIR)/test_vectors_$(CORE_SIZE).o $(CORE_OBJ) $(addprefix $(OBJ_DIR)/,$(PLATFORM_OBJ))

BINUTILS_DIR=#$(TOP)/../../externals/binutils-gdb/bin/bin/

CC        = $($(ARCH)_PREFIX)gcc
AS        = $(BINUTILS_DIR)$($(ARCH)_PREFIX)as
NM        = $(BINUTILS_DIR)$($(ARCH)_PREFIX)nm
SIZE      = $(BINUTILS_DIR)$($(ARCH)_PREFIX)size
OBJCOPY   = $(BINUTILS_DIR)$($(ARCH)_PREFIX)objcopy
FLASHCMD  = avrdude -p $(call lc,$(MCU)) -P $(PROG_PORT) -c $(PROGRAMMER) -U flash:w:$(1)
RUNCMD    = $(SIMULATOR) -m $(call lc,$(MCU)) -f 16000000 -i $1
TO_REPORT = ./to_report.sh

OPT       = -Os
CSTD      = -std=gnu99

CFLAGS    = -Wall -pedantic -Wextra -gdwarf-2 $(CSTD) $(OPT) $($(ARCH)_FLAGS) \
 $(if $(SIMUL), -DSIMAVR=1 ) \
 $(USERFLAGS) $(foreach def,$(DEF), -D$(def)) \
 $(foreach f,$(shell echo $(AFLAGS) | sed 's/\(-[[:alnum:]-]\+\)[[:space:]]\+\([^-][^[:space:]]*\)/\1,\2/g'), -Wa,$(f)) \
 -Xassembler -no-pad-sections

LDFLAGS := $($(ARCH)_LD_FLAGS)

DBG_PROG=$(PROG:%.elf=%.dbg.elf)
DBG_CFLAGS = $(CFLAGS) -DDEBUG=1 -DPRINT_IMV=1
DBG_OBJ = $(OBJ:%.o=%.dbg.o)

$(subst .,:,$(PROG))_DEP=$(OBJ)
$(subst .,_,$(DBG_PROG))_DEP=$(DBG_OBJ)

all: $(PROG)

$(info obj = $(OBJ))

none:

.phony: dbg
dbg: $(DBG_PROG)

.phony: clean
clean:
	rm -f $(PROG) $(OBJ) $(DBG_PROG) $(DBG_OBJ)
	rm -f $(OBJ:.o=.lst) $(DBG_OBJ:.o=.lst)
	rm -f $(PROG:.elf=.hex) $(PROG:.elf=.map)

.pony: run
run: $(PROG)
	./$<

.pony: dbgtest
dbgtest: clean $(DBG_PROG)
	./$(DBG_PROG) $(QUICK_TEST_FILE)

.pony: test
test: $(PROG)
	./$< $(TEST_VECTOR_FILE)

.phony: measure
measure: report

.phony: size
size: $(DBG_OBJ)
	$(SIZE) $^

.phony: report
report: $(CORE_OBJ)
	$(NM) --radix d --format posix -A $^ 
	$(SIZE) -t $^
#	| $(TO_REPORT) report_$(call lc,$(ARCH))_ .txt | tail -n 6

.phony: dbgflash
dbgflash: $(DBG_PROG:%.elf=%.hex)
	$(call FLASHCMD,$<)

.phony: flash
flash: $(PROG:%.elf=%.hex)
	$(call FLASHCMD,$<)

.phony: dbgrun
dbgrun: $(DBG_PROG) report
	$(call RUNCMD, $<)

.phony: benchrun
benchrun: $(PROG) report
	$(call RUNCMD, $<)

.phony: build
build: $(PROG)

#$(PROG): $(OBJ)
#	$(Q)$(CC) $(LDFLAGS) -o $@ $^

$(PROG): $(OBJ)
	$(Q)$(CC) $(LDFLAGS) $(CFLAGS) -o $@ $^	

$(OBJ_DIR)/%.o: %.c
	$(Q)$(CC) $(CFLAGS) -c -o $@ $^

$(OBJ_DIR)/%.o: $(COMMON_SRC_DIR)/%.c
	$(Q)$(CC) $(CFLAGS) -c -o $@ $^

$(OBJ_DIR)/%.o: %.S
	$(Q)$(CC) $(CFLAGS) -c -o $@ $^ 

$(OBJ_DIR)/%.o: $(COMMON_SRC_DIR)/%.S
	$(Q)$(CC) $(CFLAGS) -c -o $@ $^

%.o: %.c
	$(Q)$(CC) $(CFLAGS) -c -o $@ $^

%.o: %.S
	$(Q)$(CC) $(CFLAGS) -c -o $@ $^ 


$(DBG_PROG): $(DBG_OBJ)
	$(Q)$(CC) $(LDFLAGS) $(CFLAGS) -o $@ $^ 

$(OBJ_DIR)/%.dbg.o: %.c
	$(Q)$(CC) $(DBG_CFLAGS) -c -o $@ $^

$(OBJ_DIR)/%.dbg.o: %.S
	$(Q)$(CC) $(DBG_CFLAGS) -c -o $@ $^

$(OBJ_DIR)/%.dbg.o: $(COMMON_SRC_DIR)/%.c
	$(Q)$(CC) $(DBG_CFLAGS) -c -o $@ $^

$(OBJ_DIR)/%.dbg.o: $(COMMON_SRC_DIR)/%.S
	$(Q)$(CC) $(DBG_CFLAGS) -c -o $@ $^

%.dbg.o: %.c
	$(Q)$(CC) $(DBG_CFLAGS) -c -o $@ $^

%.dbg.o: %.S
	$(Q)$(CC) $(DBG_CFLAGS) -c -o $@ $^

%.hex: %.elf
	@$(OBJCOPY) -O ihex $^ $@
	
test_vectors_%.inc:  test_vectors_%.txt
	$(TOP)/common/conv.sh $^ > $@

