/* performance_test.c */
/*
    Copyright (C) 2006-2015 bg nerilex (bg@nerilex.org)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * author: bg nerilex
 * email:  bg@nerilex.org
 * license: GPLv3
 *
 *
 **/

#include "config.h"
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include "performance_test.h"

#ifndef TIMSK
#define TIMSK TIMSK1
#endif

#if SIMAVR

#ifdef ATMEGA128
#define GPIOR0 EEARL
#define GPIOR1 EEARH
#endif

void simavr_set_label_P(PGM_P s)
{
    char c;
    while (((c = pgm_read_byte(s)))) {
        GPIOR0 = c;
        s = &((char *)s)[1];
    }
}

#if 0

void calibrateTimer(void){
    (void)const_overhead;
    (void)int_overhead;
}

void startTimer(uint8_t granularity){
    GPIOR0 = 0;
}

uint32_t stopTimer(void){
    GPIOR0 = 1;
    return 0;
}

void getOverhead(uint16_t *constoh, uint16_t *intoh){
    *constoh = 0;
    *intoh   = 0;
}

void print_time_P(PGM_P s, uint32_t t){
    (void)s;
    (void)t;
}

void print_overhead(void){
}
#endif

#else

static volatile uint32_t ovfcounter;

static uint16_t const_overhead = 0;
static uint16_t int_overhead = 0;

ISR(TIMER1_OVF_vect)
{
    ++ovfcounter;
}

void calibrateTimer(void)
{
    cli();
    startTimer(1);
    const_overhead = TCNT1;
    stopTimer();
    sei();
    startTimer(1);
    TCNT1 = 0xFFFE;
    asm("nop");
    asm("nop");
    asm("nop");
    int_overhead = TCNT1;
    stopTimer();
}

void startTimer(uint8_t granularity)
{
    TCCR1B = 0; /* stop timer */
    TCNT1 = 0;
    ovfcounter = 0;
    TCCR1A = 0x00;
    TIMSK = _BV(TOIE1);         /* enable TOIE1 */
    TCCR1B = granularity & 0x7; /* start timer */
}

uint32_t stopTimer(void)
{
    uint16_t t = TCNT1;
    TCCR1B = 0; /* stop timer */
    uint64_t ret;
    ret = (((uint64_t)ovfcounter) << 16) | t;
    ret -= const_overhead;
    ret -= ovfcounter * int_overhead;
    return ret;
}

void getOverhead(uint16_t *constoh, uint16_t *intoh)
{
    *constoh = const_overhead;
    *intoh = int_overhead;
}

void print_time_P(PGM_P s, uint32_t t)
{
    printf_P(PSTR("%S%11" PRIu32), s, t);
}

void print_overhead(void)
{
    printf_P(PSTR("\n=== benchmark ===\n\tconst overhead:     %7" PRIu16
                  "\n\tinterrupt overhead: %7" PRIu16 "\n"),
             const_overhead, int_overhead);
}

#endif
