/* gage_core_232.S */
/*
    This file is part of lwc-gage-avr.
    Copyright (C) 2019 bg nerilex (bg@nerilex.org)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

.nolist
#include <avr/io.h>
#include "parameters.h"
#include "avr-asm-macros.S"
.list

STATE_B = INTERNAL_STATE_SIZE_BYTES
BS_STATE_B = 32

/* we earlier used the following form of bitsliced state:
 *
 *  0x00 : R0_0 R0_1 R0_2 R0_3
 *  0x04 : R1_0 R1_1 R1_2 R1_3
 *  0x08 : R2_0 R2_1 R2_2 R2_3
 *  0x0c : R3_0 R3_1 R3_2 R3_3
 *  0x10 : R4_0 R4_1 R4_2 R4_3
 *  0x14 : R5_0 R5_1 R5_2 R5_3
 *  0x18 : R6_0 R6_1 R6_2 R6_3
 *  0x1c : R7_0 R7_1 R7_2 R7_3
 *
 * now we rearrange the bytes in the following way:
 *
 *  0x00 : R0_3 R1_3 R2_3 R3_3 R4_3 R5_3 R6_3 R7_3
 *  0x08 : R0_2 R1_2 R2_2 R3_2 R4_2 R5_2 R6_2 R7_2
 *  0x10 : R0_1 R1_1 R2_1 R3_1 R4_1 R5_1 R6_1 R7_1
 *  0x18 : R0_0 R1_0 R2_0 R3_0 R4_0 R5_0 R6_0 R7_0
 *
 * R0_3 is the most significant byte of the 32-bit word which holds the
 * most significant bits of the left padded original state
 *
 *
 */

tmp_a = 0
tmp_b = 20
ctr3 = 21
ctr = 18
ctr2 = 19

param = 18

t0 = 20
t1 = 21
t2 = 22
t3 = 23
t4 = 24
t5 = 25
t6 = 26
t7 = 27

tmp_0 = t4
tmp_1 = t5
tmp_2 = t6
tmp_3 = t7

q1 = tmp_2
q2 = tmp_3
foo = q2
a = t0
b = t1
c = t2
d = t3

isrc = t0
idst = t1
base = t2
base_l = t2
base_h = t3

Xl = 26
Xh = 27
Zl = 30
Zh = 31

/*
 * void convert_to_bitslice(struct bs_state *bs, const struct state *s);
 */
/*
 * this is mostly like the first shuffle_function:
 * fetch a byte and put its bits into the temporary buffer
 */

 /* the MSB of s.a[0] will end up as the effective MSB of R0
  * the LSB of s.a[29] will end up as the LSB of R7
  */

/******************************************************************************/

convert_to_bitslice:
; copying src to dest first, than make an inplace conversion
#	movw Z, r24
	; set Z to the start of the state
; using t0..t7, Z, ctr, ctr2, tmp_a
	ldi ctr, BS_STATE_B / 8
1:
	ldi ctr2, 8
2:
	ld tmp_a, Z+
	.irp bs_reg, t0, t1, t2, t3, t4, t5, t6, t7
		lsl tmp_a
		rol \bs_reg
	.endr
	dec ctr2
	brne 2b
	sbiw Z, 8
	.irp bs_reg, t0, t1, t2, t3, t4, t5, t6, t7
		st Z+, \bs_reg
	.endr
	dec ctr
	brne 1b
	sbiw Z, BS_STATE_B
	ret

/******************************************************************************/

convert_from_bitslice:
	ldi ctr, BS_STATE_B / 8
1:
	.irp index, 0, 1, 2, 3, 4, 5, 6, 7
		ldd t\index, Z+(\index)
	.endr

	ldi ctr2, 8
2:
	.irp index, 0, 1, 2, 3, 4, 5, 6, 7
		lsl t\index
		rol tmp_a
	.endr
	st Z+, tmp_a
	dec ctr2
	brne 2b
	dec ctr
	brne 1b
	sbiw Z, BS_STATE_B
	ret

/******************************************************************************/
/*
 * void shuffle_state_bits_bs(struct bs_state *s);
 */

.macro load_bsreg_t bsreg:req
	ldd t0, Z+(\bsreg + 8 * 3)
	ldd t1, Z+(\bsreg + 8 * 2)
	ldd t2, Z+(\bsreg + 8 * 1)
	ldd t3, Z+(\bsreg + 8 * 0)
.endm

.macro store_bsreg_t bsreg:req
	std Z+(\bsreg +  8 * 3), t0
	std Z+(\bsreg +  8 * 2), t1
	std Z+(\bsreg +  8 * 1), t2
	std Z+(\bsreg +  8 * 0), t3
.endm

.macro load_bsreg_tmp bsreg:req
	ldd tmp_0, Z+(\bsreg + 8 * 3)
	ldd tmp_1, Z+(\bsreg + 8 * 2)
	ldd tmp_2, Z+(\bsreg + 8 * 1)
	ldd tmp_3, Z+(\bsreg + 8 * 0)
.endm

.macro store_bsreg_tmp bsreg:req
	std Z+(\bsreg +  8 * 3), tmp_0
	std Z+(\bsreg +  8 * 2), tmp_1
	std Z+(\bsreg +  8 * 1), tmp_2
	std Z+(\bsreg +  8 * 0), tmp_3
.endm


.macro bs_rotl_RR store_reg, load_reg, rotate
	load_bsreg_t ( \load_reg )
	ldi param, \rotate
	rcall rotl32
	store_bsreg_t ( \store_reg )
.endm

; rotate the bits in (t0,t1,t2,t3) to the left by n>0 bit, n is stored in param
rotl32:
1:
	lsl t0
	rol t1
	rol t2
	rol t3
	adc t0, r1 ; add zero and carry to insert carry as LSB
	dec param
	brne 1b
	ret

shuffle_state_bits_bs:
;    R0 = rotl(R0, 7);
	bs_rotl_RR 0, 0, 7
;    R5 = rotl(R5, 2);
	bs_rotl_RR 5, 5, 2
;    t = R1;
	ldd tmp_0, Z+(3 * 8 + 1)
	ldd tmp_1, Z+(2 * 8 + 1)
	ldd tmp_2, Z+(1 * 8 + 1)
	ldd tmp_3, Z+(0 * 8 + 1)
;    R1 = rotl(R2, 5);
	bs_rotl_RR 1, 2, 5
;    R2 = rotl(R4, 3);
	bs_rotl_RR 2, 4, 3
;    R4 = rotl(R6, 1);
	bs_rotl_RR 4, 6, 1
;    R6 = rotl(t, 6);
	movw t0, tmp_0
	movw t2, tmp_2
	ldi param, 6
	rcall rotl32
	store_bsreg_t ( 6 )

;    t = R7;
	load_bsreg_tmp(7)
;    R7 = rotl(R3, 4);
	bs_rotl_RR 7, 3, 4
;    R3 = t;
	store_bsreg_tmp(3)
	ret

/******************************************************************************/
/*
 dtransform_bs(struct bs_state *s, const uint8_t l)
*/

/*
foo: Qt(a,b,c,d) ( (a) ^ ((b) & ((c) ^ (d))))
q1:  Q1(a,b,c,d) ((c) ^ foo )
q2: Q2(a,b,c,d)  ( ~ ((b) ^ (d) ^ foo ) )
*/

Q:
; BEGIN compute foo
	mov foo, c
	eor foo, d
	and foo, b
	eor foo, a
; END compute foo
	mov q1, c
	eor q1, foo
	; mov q2, foo ; q2 is an alias for foo
	eor q2, d
	eor q2, b
	com q2
	ret

dtransform_bs:
; fill r0 with auxiliary "leaders" (carrys from R6 and R7 to R0 and R1)
	adiw Z, 6 ; Z points to MSB of R6
	ldi ctr, 3
1:
	ld a, Z
	ror a
	rol r0
	ldd a, Z+(1)
	ror a
	rol r0
	adiw Z, 8
	dec ctr
	brne 1b
; Z is now at LSB of R6

	sbiw Z, (6 + 3 * 8 - (3 * 8 + 4)); LSB of R4

	ldi ctr2, 4
1:
	ldi ctr, 3

	ldd d, Z+(3 + 0)
	mov tmp_1, d
	ror r0
	ror tmp_1

	ldd c, Z+(2 + 0)
	mov tmp_0, c
	ror r0
	ror tmp_0
2:
	; in this inner loop we iterate over the bitslice registers, a byte at a time
	; Z: R4, R2, R0
	ldd a, Z+(0)
	ldd b, Z+(1)
	rcall Q
	std Z+(2), q1
	std Z+(3), q2
	movw c, a
	sbiw Z, 2 ; switch from R(n) to R(n-2)
	dec ctr
	brne 2b

; now we are at R(-2)
; R(-2) := l>>1 || R(6)>>1  // tmp_0
; R(-1) := l&1  || R(7)>>1  // tmp_1
	movw a, tmp_0
	cpi ctr2, 1
	brne 3f
3:
	rcall Q
	std Z+(2), q1
	std Z+(3), q2

	sbiw Z, 1 * 8 - 6; back to R(4), but one byte more significant
	dec ctr2
	brne 1b

	adiw Z, 1 * 8 - 4
	ret

/******************************************************************************/

.global gage_core_256
.global qpermutation

gage_core_256:
	ldi r22, ROUNDS
; next comes the start of qpermutation, so we can skip the call (or jump)


s_backup = 14
s_backup_lo = 14
s_backup_hi = 15
round = 16
lfsr = 17
; s in      r24:r25
; rounds in r22

; NOTE: LFSR computation works only for an EVEN NUMBER of ROUNDS!
qpermutation:
; we need to backup some registers
; all functions we call require the state pointer in Z
; and on exit will leave the state pointer in Z
	push_ lfsr, round

	mov round, r22
	movw Z, r24
	rcall convert_to_bitslice
	ldi lfsr, 0xAA
	rjmp 3f
1:
	rcall shuffle_state_bits_bs
3:
	clr r24
	ror lfsr
	sbc r24, r24 ; this should not modify the carry; r24 = carry ? 0xff : 0x00
	rol r0
	lsl r0
	bst round, 0
	bld r0, 0
	andi r24, 0xB8
	eor lfsr, r24
	rcall dtransform_bs

	dec round
	brne 1b
99:
	pop_ round, lfsr
	rjmp convert_from_bitslice
	ret

